/*
 * This file is part of pyfreespace.
 *
 * Copyright (c) 2009 Hillcrest Laboratories, Inc.
 *
 * pyfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "freespacedeviceobject.h"
#include "freespacemoduleasync.h"
#include "freespacemodulemessages.h"
#include "structmember.h"

static void
FreespaceDeviceObject_dealloc(FreespaceDeviceObject* self)
{
    self->ob_type->tp_free((PyObject*)self);
}

static char FreespaceDevice_flush__doc__[] = "Flush all of the messages out of any receive queues.";
static PyObject *
FreespaceDevice_flush(FreespaceDeviceObject *self, PyObject *args)
{
    int rc = freespace_flush(self->freespaceDevice);
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error in flush. Freespace return code: %d", rc);
    }
    Py_INCREF(Py_None);
    return Py_None;

}

static char FreespaceDevice_openDevice__doc__[] = "Open a Freespace device for use.";
static PyObject *
FreespaceDevice_openDevice(FreespaceDeviceObject *self, PyObject *args)
{
    int rc = freespace_openDevice(self->freespaceDevice);
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error opening device. Freespace return code: %d", rc);
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static char FreespaceDevice_closeDevice__doc__[] = "Close a Freespace device.";
static PyObject *
FreespaceDevice_closeDevice(FreespaceDeviceObject *self, PyObject *args)
{
    freespace_closeDevice(self->freespaceDevice);
    Py_INCREF(Py_None);
    return Py_None;
}

static char FreespaceDevice_sendMessage__doc__[] = "Send a message to the specified Freespace device synchronously.";
static PyObject *
FreespaceDevice_sendMessage(FreespaceDeviceObject *self, PyObject *args)
{
    uint8_t message[FREESPACE_MAX_OUTPUT_MESSAGE_SIZE];
    int length;
    int rc;
    PyObject *messageObject;

    if (!PyArg_ParseTuple(args, "O:sendmessage", &messageObject)) {
        PyErr_SetString(PyExc_TypeError, "freespace.FreespaceDevice.send_message requires a message object");
        return NULL;
    }

    length = encodeMessage(messageObject, message, sizeof(message));
    if (length <= 0) {
        PyErr_SetString(PyExc_RuntimeError, "Error encoding message.");
        return NULL;
    }
    rc = freespace_send(self->freespaceDevice, message, length);
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error sending message. Freespace return code: %d", rc);
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static char FreespaceDevice_sendMessageAsync__doc__[] = "Send a message to the specified Freespace device, but do not block.";
static PyObject *
FreespaceDevice_sendMessageAsync(FreespaceDeviceObject *self, PyObject *args, PyObject* keywds)
{
    uint8_t message[FREESPACE_MAX_OUTPUT_MESSAGE_SIZE];
    int length;
    int rc;
    PyObject *messageObject;
    PyObject *callback = NULL;
    PyObject *cookie = NULL;
    int timeout;

    static char *kwlist[] = {"message", "timout", "callback", NULL};

    /*
     * Method takes a message object and a timeout
     * And optionally a callback function and data to pass to the callback function
     */
    if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oi|O:sendmessage_async", kwlist,
                                     &messageObject, &timeout, &callback)) {
        return NULL;
    }
    if (callback != NULL) {
        if (!PyCallable_Check(callback)) {
            PyErr_SetString(PyExc_TypeError, "callback parameter must be callable");
            return NULL;
        }
    }

    Py_XINCREF(callback);         /* Add a reference to new callback */
    Py_XDECREF(pySendCallback_);  /* Dispose of previous callback */
    pySendCallback_ = callback;   /* Remember new callback */

    length = encodeMessage(messageObject, message, sizeof(message));
    if (length <= 0) {
        PyErr_SetString(PyExc_RuntimeError, "Error encoding message.");
        return NULL;
    }

    rc = freespace_sendAsync(self->freespaceDevice, message, length, timeout, sendCallback, (void*) self);
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error sending message. Freespace return code: %d", rc);
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static char FreespaceDevice_readMessage__doc__[] =
"Read a message from the specified device.  This function will block until a message is received,\
 the timout is reached, or there's an error.\
 Return the appropriate message object or None if the message cannot be decoded.";
static PyObject *
FreespaceDevice_readMessage(FreespaceDeviceObject *self, PyObject *args)
{
    int timeout, actualLen, rc;
    uint8_t buffer[FREESPACE_MAX_INPUT_MESSAGE_SIZE];

    if (!PyArg_ParseTuple(args, "i:readmessage", &timeout)) {
        PyErr_SetString(PyExc_TypeError, "freespace.FreespaceDevice.read_message requires an integer timeout");
        return NULL;
    }

    rc = freespace_read(self->freespaceDevice, buffer, sizeof(buffer), timeout, &actualLen);
    if (rc == FREESPACE_ERROR_TIMEOUT) {
        Py_INCREF(Py_None);
        return Py_None;
    }
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error reading message. Freespace return code: %d", rc);
    }

    return decodeMessage(buffer, actualLen);
}

static struct PyMethodDef FreespaceDevice_methods[] = {
    {"open",                (PyCFunction)FreespaceDevice_openDevice,    METH_NOARGS,    FreespaceDevice_openDevice__doc__},
    {"sendmessage",         (PyCFunction)FreespaceDevice_sendMessage,   METH_VARARGS,   FreespaceDevice_sendMessage__doc__},
    {"sendmessage_async",   (PyCFunction)FreespaceDevice_sendMessageAsync,   METH_VARARGS | METH_KEYWORDS,   FreespaceDevice_sendMessageAsync__doc__},
    {"readmessage",         (PyCFunction)FreespaceDevice_readMessage,   METH_VARARGS,   FreespaceDevice_readMessage__doc__},
    {"close",               (PyCFunction)FreespaceDevice_closeDevice,   METH_NOARGS,    FreespaceDevice_closeDevice__doc__},
    {"flush",               (PyCFunction)FreespaceDevice_flush,   METH_NOARGS,    FreespaceDevice_flush__doc__},
    {"set_receive_callback",(PyCFunction)fs_set_receive_callback, METH_VARARGS, FreespaceDevice_set_receive_callback__doc__},

    {NULL,  NULL}  /* sentinel */
};

static struct PyMemberDef FreespaceDevice_members[] = {
    {"vendor", T_INT, offsetof(FreespaceDeviceObject, vendor), READONLY, "Device vendor ID code"},
    {"product", T_INT, offsetof(FreespaceDeviceObject, product), READONLY, "Device product ID"},
    {"name", T_OBJECT, offsetof(FreespaceDeviceObject, name), READONLY, "The user-meaningful name for the device"},
    {"id", T_INT, offsetof(FreespaceDeviceObject, freespaceDevice), READONLY, "The unique ID for this device"},
    {NULL}  /* Sentinel */
};


/* ---------- */
/* Generic type method implementations (e.g. __str__, __compare__, etc.) */

static PyObject *
FreespaceDevice_str(FreespaceDeviceObject * obj)
{
    return PyString_FromFormat("Device = %s\n   Vendor ID  = 0x%x (%d)\n   Product ID = 0x%x (%d)\n",
            PyString_AsString(obj->name), obj->vendor, obj->vendor, obj->product, obj->product);
}

static int
FreespaceDevice_compare(FreespaceDeviceObject * obj1, FreespaceDeviceObject * obj2)
{
    int result;

    if (obj1->freespaceDevice < obj2->freespaceDevice) {
        result = -1;
    } else if (obj1->freespaceDevice > obj2->freespaceDevice) {
        result = 1;
    } else {
        result = 0;
    }
    return result;
}

/* ---------- */

static char FreespaceDeviceType__doc__[] = "Opaque handle to FreespaceDevice instances.";
static PyTypeObject FreespaceDeviceType = {
    PyObject_HEAD_INIT(NULL)
    0,                                          /*ob_size*/
    "freespace.FreespaceDevice",                /*tp_name*/
    sizeof(FreespaceDeviceObject),              /*tp_basicsize*/
    0,                                          /*tp_itemsize*/
    /* methods */
    (destructor) FreespaceDeviceObject_dealloc, /*tp_dealloc*/
    0,                                          /*tp_print*/
    0,                                          /*tp_getattr*/
    0,                                          /*tp_setattr*/
    (cmpfunc) FreespaceDevice_compare,          /*tp_compare*/
    0,                                          /*tp_repr*/
    0,                                          /*tp_as_number*/
    0,                                          /*tp_as_sequence*/
    0,                                          /*tp_as_mapping*/
    0,                                          /*tp_hash*/
    0,                                          /*tp_call*/
    (reprfunc) FreespaceDevice_str,             /*tp_str*/
    0,                                          /*tp_getattro*/
    0,                                          /*tp_setattro*/
    0,                                          /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT,                         /*tp_flags*/
    FreespaceDeviceType__doc__,                 /* Documentation string */
    0,                                          /* tp_traverse */
    0,                                          /* tp_clear */
    0,                                          /* tp_richcompare */
    0,                                          /* tp_weaklistoffset */
    0,                                          /* tp_iter */
    0,                                          /* tp_iternext */
    FreespaceDevice_methods,                    /* tp_methods */
    FreespaceDevice_members,                    /* tp_members */
    0,                                          /* tp_getset */
    0,                                          /* tp_base */
    0,                                          /* tp_dict */
    0,                                          /* tp_descr_get */
    0,                                          /* tp_descr_set */
    0,                                          /* tp_dictoffset */
    0,                                          /* tp_init */
    0,                                          /* tp_alloc */
    0,                                          /* tp_new */
};

/* End of code for FreespaceDevice objects */
/* -------------------------------------------------------- */


PyObject* createDeviceObject(FreespaceDeviceId id) {
    FreespaceDeviceObject* dev;
    struct FreespaceDeviceInfo info;
    int rc;

    dev = (FreespaceDeviceObject*)FreespaceDeviceType.tp_alloc(&FreespaceDeviceType, 0);
    if (dev != NULL) {
        dev->freespaceDevice = id;
        rc = freespace_getDeviceInfo(dev->freespaceDevice, &info);
        if (rc != FREESPACE_SUCCESS) {
            Py_DECREF(dev);
            return NULL;
        }
        dev->name = PyString_FromString(info.name);
        dev->product = info.product;
        dev->vendor = info.vendor;
    }

    return (PyObject*) dev;
}


void addDeviceObject(PyObject* m) {
    if (PyType_Ready(&FreespaceDeviceType) < 0)
        return;
    Py_INCREF(&FreespaceDeviceType);
    PyModule_AddObject(m, "FreespaceDevice", (PyObject *)&FreespaceDeviceType);
}

