/*
 * This file is part of pyfreespace.
 *
 * Copyright (c) 2009 Hillcrest Laboratories, Inc.
 *
 * pyfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "Python.h"
#include "freespace/freespace.h"
#include "freespacemodulemessages.h"
#include "freespacemoduleasync.h"
#include "freespacedeviceobject.h"

static PyObject *ErrorObject;


static char fs_freespace_init__doc__[] = " Initialize the Freespace library. return 0 on success.";
static PyObject *
fs_freespace_init(PyObject *self /* Not used */, PyObject *args)
{
    int rc;

    rc = freespace_init();
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error initializing freespace. Freespace return code: %d", rc);
    }

    // Set up libfreespace to call into all of our callbacks.
    // We'll only pass on the callbacks once the python user has registered a python function to call
    freespace_setFileDescriptorCallbacks(add_pollfd, remove_pollfd);
    freespace_setDeviceHotplugCallback(hotplugCallback, NULL);

    Py_INCREF(Py_None);
    return Py_None;
}

static char fs_freespace_exit__doc__[] = "Finalize the Freespace library.";
static PyObject *
fs_freespace_exit(PyObject *self /* Not used */, PyObject *args)
{
    freespace_exit();
    Py_INCREF(Py_None);
    return Py_None;
}

static char fs_freespace_getdevices__doc__[] = "Get a list of available Freespace Devices.";
static PyObject *
fs_freepace_getDevices(PyObject *self /* Not used */, PyObject *args) {
    FreespaceDeviceObject* dev;
    PyObject* list;
    FreespaceDeviceId deviceIds[FREESPACE_MAXIMUM_DEVICE_COUNT];
    int numIds;
    int rc;
    int i;

    list = PyList_New(0);
    if (list == NULL) {
        return NULL;
    }

    rc = freespace_getDeviceList(deviceIds, FREESPACE_MAXIMUM_DEVICE_COUNT, &numIds);
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error getting device list. Freespace return code: %d", rc);
    }
    for (i = 0; i < numIds; i++) {
        dev = (FreespaceDeviceObject*) createDeviceObject(deviceIds[i]);
        if (dev != NULL) {
            rc = PyList_Append(list, (PyObject*) dev);
            if (rc < 0) {
                Py_DECREF(dev);
                Py_DECREF(list);
                return NULL;
            }
        }
    }

    return list;
}

/* List of methods defined in the module */
static struct PyMethodDef fs_methods[] = {
    {"init", (PyCFunction)fs_freespace_init, METH_NOARGS, fs_freespace_init__doc__},
    {"exit", (PyCFunction)fs_freespace_exit, METH_NOARGS, fs_freespace_exit__doc__},
    {"get_device_list", (PyCFunction)fs_freepace_getDevices, METH_NOARGS, fs_freespace_exit__doc__},
    {"perform", (PyCFunction)fs_perform, METH_NOARGS, fs_freespace_perform__doc__},
    {"set_filedescriptor_callbacks", (PyCFunction)fs_set_fd_callbacks, METH_VARARGS | METH_KEYWORDS, fs_set_fd_callbacks__doc__},
    {"sync_filedescriptors", (PyCFunction)fs_sync_fds, METH_NOARGS, fs_sync_fds__doc__},
    {"get_next_timeout", (PyCFunction)fs_get_next_timeout, METH_NOARGS, fs_getNextTimeout__doc__},
    {"set_hotplug_callback", (PyCFunction) fs_set_hotplug_callback, METH_VARARGS, fs_set_hotplug_callback__doc__},

    {NULL,  (PyCFunction)NULL, 0, NULL}  /* sentinel */
};


static char freespace_module_documentation[] =
""
;

/* Initialization function for the module (*must* be called initfreespace) */
PyMODINIT_FUNC initfreespace() {
    PyObject *m, *d;


    /* Create the module and add the functions */
    m = Py_InitModule4("freespace", fs_methods,
                       freespace_module_documentation,
                       (PyObject*)NULL,PYTHON_API_VERSION);

    /* Add some symbolic constants to the module */
    d = PyModule_GetDict(m);
    ErrorObject = PyString_FromString("freespace.error");
    PyDict_SetItemString(d, "error", ErrorObject);

    addDeviceObject(m);

    addMessageObjects(m);


    PyModule_AddIntMacro(m, FREESPACE_HOTPLUG_INSERTION);
    PyModule_AddIntMacro(m, FREESPACE_HOTPLUG_REMOVAL);

    PyModule_AddIntMacro(m, FREESPACE_SUCCESS);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_IO);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_ACCESS);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_NO_DEVICE);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_NOT_FOUND);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_BUSY);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_TIMEOUT);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_PIPE);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_INTERRUPTED);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_OUT_OF_MEMORY);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_SEND_TOO_LARGE);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_INVALID_DEVICE);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_RECEIVE_BUFFER_TOO_SMALL);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_COULD_NOT_CREATE_THREAD);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_BUFFER_TOO_SMALL);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_NO_DATA);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_MALFORMED_MESSAGE);
    PyModule_AddIntMacro(m, FREESPACE_ERROR_UNEXPECTED);


    /* Check for errors */
    if (PyErr_Occurred())
        Py_FatalError("can't initialize module freespace");
}

