/*
 * This file is part of pyfreespace.
 *
 * Copyright (c) 2009 Hillcrest Laboratories, Inc.
 *
 * pyfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include "freespacemoduleasync.h"
#include "freespacemodulemessages.h"
#include "freespacedeviceobject.h"


/* --------------- Message send/receive callbacks ----------------- */

void sendCallback(FreespaceDeviceId id, void* cookie, int rc) {
    PyObject *arglist;
    PyObject *result;
    PyObject *self;

    // If there's no callback into python, don't do anything
    if (pySendCallback_ == NULL) {
        return;
    }

    if (cookie == NULL) {
        return;
    }

    self = (PyObject*) cookie;

    /* Time to call the callback */
    arglist = Py_BuildValue("(Oi)", self, rc);
    if (arglist == NULL) {
        return;
    }

    result = PyObject_CallObject(pySendCallback_, arglist);

    Py_DECREF(arglist);
    Py_XDECREF(result);
}

void receiveCallback(FreespaceDeviceId id, const uint8_t* message, int length, void* cookie, int rc) {
    PyObject *arglist;
    PyObject *result;
    PyObject* messageObject = NULL;
    PyObject *self;

    if (cookie == NULL) {
        return;
    }
    self = (PyObject*) cookie;

    if (pyReceiveCallback_ == NULL) {
        return;
    }

    if (rc == FREESPACE_SUCCESS) {
        messageObject = decodeMessage(message, length);
    } else {
        Py_INCREF(Py_None);
        messageObject = Py_None;
    }

    if (messageObject == NULL) {
        return;
    }

    /* Time to call the callback */
    arglist = Py_BuildValue("(OOi)", self, messageObject, rc);
    if (arglist == NULL) {
        return;
    }

    result = PyObject_CallObject(pyReceiveCallback_, arglist);

    Py_DECREF(arglist);
    Py_XDECREF(result);
}

PyObject* fs_set_receive_callback(FreespaceDeviceObject *self, PyObject *args) {
    PyObject* pyreceiveCallback = NULL;

    if (!PyArg_ParseTuple(args, "O", &pyreceiveCallback)) {
        return NULL;
    }
    if (pyreceiveCallback != NULL) {
        if (!PyCallable_Check(pyreceiveCallback)) {
            PyErr_SetString(PyExc_TypeError, "callback parameter must be a callable");
            return NULL;
        }
    }

    Py_XINCREF(pyreceiveCallback);         /* Add a reference to new callback */
    Py_XDECREF(pyReceiveCallback_);  /* Dispose of previous callback */
    pyReceiveCallback_ = pyreceiveCallback;

    freespace_setReceiveCallback(self->freespaceDevice, receiveCallback, self);

    Py_INCREF(Py_None);
    return Py_None;
}

/* --------------- FD add/remove callbacks ----------------- */

static PyObject* pyAddFDCallback_ = NULL;
static PyObject* pyRemoveFDCallback_ = NULL;

void add_pollfd(FreespaceFileHandleType fd, short events) {
    PyObject *arglist;
    PyObject *result;

    if (pyAddFDCallback_ == NULL) {
        return;
    }

    /* Time to call the callback */
    arglist = Py_BuildValue("(ii)", fd, events);
    if (arglist == NULL) {
        return;
    }

    result = PyObject_CallObject(pyAddFDCallback_, arglist);

    Py_DECREF(arglist);
    Py_XDECREF(result);
}

void remove_pollfd(FreespaceFileHandleType fd) {
    PyObject *arglist;
    PyObject *result;

    if (pyRemoveFDCallback_ == NULL) {
        return;
    }

    /* Time to call the callback */
    arglist = Py_BuildValue("(i)", fd);
    if (arglist == NULL) {
        return;
    }

    result = PyObject_CallObject(pyRemoveFDCallback_, arglist);

    Py_DECREF(arglist);
    Py_XDECREF(result);
}

PyObject* fs_set_fd_callbacks(PyObject *self /* Not used */, PyObject *args) {
    PyObject* addCB = NULL;
    PyObject* remCB = NULL;

    if (!PyArg_ParseTuple(args, "OO", &addCB, &remCB)) {
        return NULL;
    }
    if (addCB != NULL) {
        if (!PyCallable_Check(addCB)) {
            PyErr_SetString(PyExc_TypeError, "callback parameters must be callables");
            return NULL;
        }
    }
    if (remCB != NULL) {
        if (!PyCallable_Check(remCB)) {
            PyErr_SetString(PyExc_TypeError, "callback parameters must be callables");
            return NULL;
        }
    }
    Py_XINCREF(addCB);         /* Add a reference to new callback */
    Py_XDECREF(pyAddFDCallback_);  /* Dispose of previous callback */
    pyAddFDCallback_ = addCB;


    Py_XINCREF(remCB);         /* Add a reference to new callback */
    Py_XDECREF(pyRemoveFDCallback_);  /* Dispose of previous callback */
    pyRemoveFDCallback_ = remCB;

    Py_IncRef(Py_None);
    return Py_None;
}

PyObject* fs_sync_fds(PyObject *self /* Not used */, PyObject *args) {
    int rc;

    rc = freespace_syncFileDescriptors();
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error syncing file descriptors. Freespace return code: %d", rc);
    }

    Py_IncRef(Py_None);
    return Py_None;
}


/* --------------- Hotplug  ----------------- */
static PyObject* pyHotplugCallback_ = NULL;

void hotplugCallback(enum freespace_hotplugEvent event, FreespaceDeviceId id, void* cookie) {
    PyObject *arglist;
    PyObject *result;
    PyObject *device;

    if (pyHotplugCallback_ == NULL) {
        return;
    }

    device = createDeviceObject(id);
    if (device == NULL) {
        return;
    }

    /* Time to call the callback */
    arglist = Py_BuildValue("(iO)", event, device);
    if (arglist == NULL) {
        return;
    }

    result = PyObject_CallObject(pyHotplugCallback_, arglist);

    Py_DECREF(arglist);
    Py_XDECREF(result);
}

PyObject* fs_set_hotplug_callback(PyObject* self /*Not used */, PyObject *args) {
    PyObject* hotplugCallback = NULL;

    if (!PyArg_ParseTuple(args, "O", &hotplugCallback)) {
        return NULL;
    }
    if (hotplugCallback != NULL) {
        if (!PyCallable_Check(hotplugCallback)) {
            PyErr_SetString(PyExc_TypeError, "callback parameters must be callables");
            return NULL;
        }
    }

    Py_XINCREF(hotplugCallback);         /* Add a reference to new callback */
    Py_XDECREF(pyHotplugCallback_);  /* Dispose of previous callback */
    pyHotplugCallback_ = hotplugCallback;

    Py_IncRef(Py_None);
    return Py_None;
}

/* --------------- Basic poll/perform stuff ----------------- */

PyObject* fs_get_next_timeout(PyObject* self /* Not used */, PyObject *args) {
    int rc;
    int timeout;

    rc = freespace_getNextTimeout(&timeout);
    if (rc != FREESPACE_SUCCESS) {
        return PyErr_Format(PyExc_RuntimeError, "Error getting next timeout. Freespace return code: %d", rc);
    }

    return PyInt_FromLong(timeout);
}

PyObject* fs_perform(PyObject *self /* Not used */, PyObject *args) {
    int rc;
    rc = freespace_perform();
    if (rc != FREESPACE_SUCCESS) {
        PyErr_Format(PyExc_RuntimeError, "Error in freespace_perform. Freespace return code: %d", rc);
        return NULL;
    }

    Py_IncRef(Py_None);
    return Py_None;
}
