#!/usr/bin/env python
#
# This file is part of pyfreespace.
#  
# Copyright (c) 2009 Hillcrest Laboratories, Inc. 
# 
# pyfreespace is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import sys
import getopt
import os

def getModuleMessagesCFile():
    moduleMessagesFile = open("freespacemodulemessages.c", "w")
    return moduleMessagesFile

def writeModuleMessageOpener(file):
    file.write('''/*
 * This file is part of pyfreespace.
 * 
 * Copyright (c) 2009 Hillcrest Laboratories, Inc. 
 *
 * pyfreespace is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "Python.h"
#include "freespacemodulemessages.h"
#include "freespace/freespace_codecs.h"
#include "freespace/freespace_printers.h"
#include "structmember.h"
''')

#------------------------------------------------------------------------------------------
#---------- Create the code for the specific message structs, members, etc. --------------

def writeMessageDeclaration(file, message):
    file.write('''

/* ----------------------------------------------------- */
/* Declarations for objects of type %(name)s */
typedef struct {
    PyObject_HEAD
'''%{'name':message.name})
    if message.hasUnReservedFields():
        file.write('''    
    struct freespace_%(name)s message;'''%{'name':message.name})
    file.write('''
} %(name)s;

static PyTypeObject %(name)sType;


#define %(name)s_Check(v)  ((v)->ob_type == &%(name)sType)
'''%{'name':message.name})

def writeMessageSingleField(file, fieldName, fieldSizeType, messageName, readwrite, docstr):
    file.write('''
    {"%(fieldname)s", %(fieldSizeType)s, offsetof(%(messageName)s, message.%(fieldname)s), %(readwrite)s, "%(docstr)s"},'''%{'fieldname':fieldName,
                                                                                                                            'fieldSizeType':fieldSizeType,
                                                                                                                            'messageName':messageName,
                                                                                                                            'readwrite':readwrite,
                                                                                                                            'docstr':docstr})

def cTypeToPythonSizeMacro(str):
    # Python doesn't have switch statements
    result = {
      'int'     :'T_INT',
      'int8_t'  :'T_BYTE',
      'uint8_t' :'T_UBYTE',
      'int16_t' :'T_SHORT',
      'uint16_t':'T_USHORT',
      'uint32_t':'T_UINT'
    }[str]
    return result

def writeMessageMembers(file, message):
    if not message.hasUnReservedFields():
        return
    file.write('''
static struct PyMemberDef %(name)s_members[] = {'''%{'name':message.name})
    for field in message.Fields:
        if (field['name'] == 'RESERVED'):
            continue
        readwrite = ""
        if (message.encode):
            readwrite = "0"
        elif (message.decode):
            readwrite = "READONLY"
        if (field.has_key('cType')):
            doc = field['name']
            if field.has_key('comment'):
                doc = field['comment'].replace("\n", "")
            writeMessageSingleField(file, field['name'], cTypeToPythonSizeMacro(field['cType']), message.name, readwrite, doc)
        else:
            array = None
            fieldSize = 'T_UBYTE'
            if field.has_key('bits'):
                fieldSize = 'T_BOOL'
                array = field['bits']
            elif field.has_key('nibbles'):
                array = field['nibbles']
            if array != None:
                for smallField in array:
                    if smallField['name'] != 'RESERVED':
                        doc = smallField['name']
                        if smallField.has_key('comment'):
                            doc = smallField['comment'].replace("\n", "")
                        writeMessageSingleField(file, smallField['name'], fieldSize, message.name, readwrite, doc)
            else:
                # Assume it's just some arbitrary amount of data, use the size for allocation
                #outHeader.write('\tunsigned char %s[%d];\n'%(field['name'], field['size']))
                pass
    file.write('''
    {NULL}  /* Sentinel */
};
''')
    
def writeMessagePrintStr(file, message):
    if not message.hasUnReservedFields():
        return
    file.write('''
static PyObject *
%(name)s_str(%(name)s * obj)
{
    char str[1024];
    freespace_print%(name)sStr(str, sizeof(str), &(obj->message));
    if (str != NULL) {
        return PyString_FromString(str);
    }
    
    return NULL;
}
'''%{'name':message.name})
    
def writeMessageTypeStruct(file, message):
    printFunc = '0'
    membersStruct = '0'
    docStr = "%s object"%message.name
    if message.hasUnReservedFields():
        printFunc = "(reprfunc)%s_str"%message.name
        membersStruct = "%s_members"%message.name
    if message.Documentation != None:
        docStr = message.Documentation.replace('\n',' ')
    file.write('''
static PyTypeObject %(name)sType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "freespace.%(name)s",      /*tp_name*/
    sizeof(%(name)s),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    0,                         /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    %(printFunc)s,   /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT,        /*tp_flags*/
    "%(docstr)s",    /* tp_doc */
    0,                              /* tp_traverse */
    0,                              /* tp_clear */
    0,                              /* tp_richcompare */
    0,                              /* tp_weaklistoffset */
    0,                              /* tp_iter */
    0,                              /* tp_iternext */
    0,                              /* tp_methods */
    %(members)s,       /* tp_members */
    0,                              /* tp_getset */
    0,                              /* tp_base */
    0,                              /* tp_dict */
    0,                              /* tp_descr_get */
    0,                              /* tp_descr_set */
    0,                              /* tp_dictoffset */
    0,                              /* tp_init */
    0,                              /* tp_alloc */
};

'''%{'name':message.name,
     'printFunc':printFunc,
     'docstr':docStr,
     'members':membersStruct})
    
def writeMessageBindings(file, message):
    writeMessageDeclaration(file, message)
    writeMessageMembers(file, message)
    writeMessagePrintStr(file, message)
    writeMessageTypeStruct(file, message)
#------------------------------------------------------------------------------------------
#------------------------------------------------------------------------------------------

def writeEncodeMessage(file, messages):
    file.write('''
int encodeMessage(PyObject* messageObject, uint8_t* message, int maxlen) {''')

    for message in messages:
        if message.encode and message.shouldGenerate:
            if message.hasUnReservedFields():
                file.write('''
    if (%(name)s_Check(messageObject)) {
        return freespace_encode%(name)s(&(((%(name)s*) messageObject)->message), message, maxlen);
    }'''%{'name':message.name})
            else:
                file.write('''
    if (%(name)s_Check(messageObject)) {
        return freespace_encode%(name)s(message, maxlen);
    }'''%{'name':message.name})
    file.write('''
    return -1;
}
''')
    
def writeDecodeMessage(file, messages):
    file.write('''
PyObject* decodeMessage(const uint8_t* message, int length) {
    struct freespace_message s;''')
    
    for message in messages:
        if not message.hasUnReservedFields() or not message.decode:
            continue
        file.write('''
    %(name)s* %(structName)s;'''%{'name':message.name, 'structName':message.structName})
    
    file.write('''
    
    freespace_decode_message(message, length,  &s);''')
    
    file.write('''
    
    switch(s.messageType) {''')
    
    for message in messages:
        if not message.hasUnReservedFields() or not message.decode:
            continue
        file.write('''
    case %(enumName)s:
        %(structName)s = (%(name)s*)%(name)sType.tp_alloc(&%(name)sType, 0);
        if (%(structName)s != NULL) {
            %(structName)s->message = s.%(structName)s;
        }
        return (PyObject*)%(structName)s;'''%{'enumName':message.enumName, 
                                                'name':message.name, 
                                                'structName':message.structName})
    file.write('''
    default:
        break;
    }
    
    Py_INCREF(Py_None);
    return Py_None;
}''')
    
def writeAddMessageObjects(file, messages):
    file.write('''
void
addMessageObjects(PyObject* m) 
{''')
    for message in messages:
        if message.shouldGenerate:
            file.write('''
    // Add %(name)s type to freespace module
    %(name)sType.tp_new = PyType_GenericNew;
    if (PyType_Ready(&%(name)sType) < 0)
        return;
    Py_INCREF(&%(name)sType);
    PyModule_AddObject(m, "%(name)s", (PyObject *)&%(name)sType);
'''%{'name':message.name})
    file.write('''}\n''')



def process(messages):
    file = getModuleMessagesCFile()
    writeModuleMessageOpener(file)
    
    for message in messages:
        if message.shouldGenerate:
            writeMessageBindings(file, message)
    
    file.write('\n/* ----------------------------------------------------- */')
    
    writeEncodeMessage(file, messages)
    writeDecodeMessage(file, messages)
    file.write('\n')
    writeAddMessageObjects(file, messages)

# ---------------------- Main function --------------------------------
# Courtesy of Guido: http://www.artima.com/weblogs/viewpost.jsp?thread=4829
class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.getopt(argv[1:], "h", ["help"])
        except getopt.error, msg:
             raise Usage(msg)
        # process options
        for o, a in opts:
            if o in ("-h", "--help"):
                print __doc__
                return 0
        # process arguments
        for arg in args:
            g = {}
            d = {}
            execfile(arg, g, d)
            messages = d['messages']
            process(messages)
    except Usage, err:
        print >>sys.stderr, err.msg
        print >>sys.stderr, "for help use --help"
        return 2
        
        
if __name__ == "__main__":
    sys.exit(main())

