#!/usr/bin/env python


import freespace
import threading
from sys import platform


lock = threading.Lock()

bodyFrames = {}
userFrames = {}

quit = False

if platform == 'win32':
    import win32event;
    rlist = []
else:
    from select import poll
    poll = poll()


def add_pollfd(fd, events):
    if platform == 'win32':
        rlist.append(fd)
    else:
        poll.register(fd, events)


def remove_pollfd(fd):
    if platform == 'win32':
        rlist.remove(fd)
    else:
        poll.unregister(fd)


def receiveCallback(device, message, rc):
    if rc == freespace.FREESPACE_SUCCESS:
        if isinstance(message, freespace.BodyFrame):
            lock.acquire()
            bodyFrames[device.id] = message
            lock.release()
        if isinstance(message, freespace.UserFrame):
            lock.acquire()
            userFrames[device.id] = message
            lock.release()
            
    else:
        if rc == freespace.FREESPACE_ERROR_NOT_FOUND:
            device.close();
            
def initDevice(device):    
    device.open()
    
    # Add it to the list of devices
    lock.acquire()
    bodyFrames[device.id] = None
    userFrames[device.id] = None
    lock.release()
    
    device.flush()
    
    d = freespace.DataMotionControl()
    d.enableBodyMotion = True
    d.enableUserPosition = True
    d.enableMouseMovement = False
    device.sendmessage(d)
    
    device.set_receive_callback(receiveCallback);

def cleanupDevice(device):
    d = freespace.DataMotionControl()
    d.enableBodyMotion = False
    d.enableUserPosition = False
    d.enableMouseMovement = True
    rc = device.sendmessage(d)

    device.close()


def hotplugCallback(event, device):
    if event == freespace.FREESPACE_HOTPLUG_REMOVAL:
        print "device removed", device.id
        device.close()
    elif event == freespace.FREESPACE_HOTPLUG_INSERTION:
        print "device added", device.id
        initDevice(device)


def getBodyFrameFromInputThread(deviceId):
    lock.acquire()
    bodyFrame = bodyFrames[deviceId]
    lock.release()
    return bodyFrame
    
def getUserFrameFromInputThread(deviceId):
    lock.acquire()
    userFrame = userFrames[deviceId]
    lock.release()
    return userFrame


def getDeviceIds():
    lock.acquire()
    keys = bodyFrames.keys()
    lock.release()
    return keys

def stopInputThread():
    global quit
    quit = True

class InputThread(threading.Thread):
    global quit
    def run (self):
        freespace.init()

        freespace.set_hotplug_callback(hotplugCallback)
        freespace.set_filedescriptor_callbacks(add_pollfd, remove_pollfd)
        freespace.sync_filedescriptors()

        freespace.perform()

        try:
            while (not quit) :
                timeoutMs = freespace.get_next_timeout();

                # Adjust the maximum timeout to allow for Ctrl-C handling
                if timeoutMs < 0 or timeoutMs > 1000 :
                    timeoutMs = 1000;

                if platform == 'win32':
                    #(outrlist, foo, bar) = select.select(rlist, [], [], timeoutMs)
                    win32event.WaitForMultipleObjects(rlist, 0, timeoutMs)
                else:
                    events = poll.poll(timeoutMs)
                    if len(events) == 0:
                        continue
                freespace.perform();
        except KeyboardInterrupt:
            print "Caught Keyboard Interrupt"
        finally:
            print "Cleaning up all devices..."

            list = freespace.get_device_list();
            for device in list:
                cleanupDevice(device)

            freespace.exit()
