#!/usr/bin/env python

# This is a volleyball game developed to demonstrate
# the ease of creating an application using pyfreespace
#
# This example uses pygame (www.pygame.org)
#
# It was developed on a Mac OS 10.5 using pygame version
# 1.9.0rc2 for python 2.6

from pygame.locals import *
import sys
import pygame
from vector import *
from button import *
from input_thread import *
import random

import freespace


# --------------  Defines    -------------------------------
SCREEN_WIDTH = 1280
SCREEN_HEIGHT = 720
GROUND_HEIGHT = 100

GROUND_LEVEL = SCREEN_HEIGHT - GROUND_HEIGHT

GRAVITY = 0.9 # just like real gravity!! Except this is in pixels per frame per frame, which isn't really anything

BLACK = pygame.Color(0,0,0)
RED = pygame.Color(255,0,0)
GREEN = pygame.Color(0,255,0)
BLUE = pygame.Color(0,0,255)
WHITE = pygame.Color(255,255,255)

MAX_SPEED = 60

drag_factor = 1
bounce_factor = 0.8

screen = pygame.display.set_mode((1280,720))

left_team_score = 0
right_team_score = 0

class State:
    INIT = 0
    CHOOSE_SIDES = 1
    PLAY_GAME = 2
    PAUSED = 3
    GAME_OVER = 4
    MENU = 5
    PRACTICE = 6
    SERVING = 7
    
class Side:
    LEFT = 0
    RIGHT = 1


state = State.INIT
# ----------------------------------------------------------

class GameObject():
    def draw(self):
        pass
    
    def update(self):
        pass


class Fence(GameObject):
    """ The net and/or fence """
    def __init__(self, width, height, top):
        self.rect = Rect((SCREEN_WIDTH/2)-(width/2), top, width, height)
        
    def draw(self):
        #rect=Rect(630, 400, 20, 260)
        screen.fill(WHITE, self.rect)
    
    
    def collide(self, ball):
        centerX = self.rect.left + (self.rect.width/2)
        if math.fabs(centerX - ball.position.x) > (ball.radius + self.rect.width/2):
            return
        
        if ball.position.y + ball.radius < self.rect.top:
            return
            
        if ball.position.y >= self.rect.top:
            # Easy case: ball bounces off flat side
            if (ball.position.x < SCREEN_WIDTH/2):
                ball.position.x = self.rect.left - ball.radius
            else:
                ball.position.x = self.rect.right + ball.radius
            ball.velocity.x *= -1
            ball.velocity.scale(bounce_factor)
            return
        
        if ball.position.x >= self.rect.left and ball.position.x <= self.rect.right:
            # Second easy case: ball bounces off top
            ball.velocity.y *= -1
            ball.position.y = self.rect.top - ball.radius
            ball.velocity.scale(bounce_factor)
            return
            
        # OK, now we have the hard case, literally, the corner case
        if ball.position.x < self.rect.left:
            # top left corner
            tl = Vector2(self.rect.left, self.rect.top)
            point_to_ball = ball.position - tl
            dist = point_to_ball.len()
            if dist > ball.radius:
                return
            point_to_ball.normalize()
            backwards = -ball.velocity
            newVel = backwards.reflect(point_to_ball)
            ball.velocity = newVel
            ball.velocity.scale(bounce_factor)
            
            point_to_ball.normalize()
            point_to_ball.scale(ball.radius)
            ball.position = tl + point_to_ball
            
        if ball.position.x > self.rect.right:
            # top right corner
            tl = Vector2(self.rect.right, self.rect.top)
            point_to_ball = ball.position - tl
            dist = point_to_ball.len()
            if dist > ball.radius:
                return
            point_to_ball.normalize()
            backwards = -ball.velocity
            newVel = backwards.reflect(point_to_ball)
            ball.velocity = newVel
            ball.velocity.scale(bounce_factor)
            
            point_to_ball.normalize()
            point_to_ball.scale(ball.radius)
            ball.position = tl + point_to_ball
            

class Ball(GameObject):
    """ The ball """

    def __init__(self, x, y, color, radius, width = 0):
        self.position = Vector2(x,y)
        self.color = color
        self.radius = radius
        self.width = width
        self.velocity = Vector2(0,0)
        
    def draw(self):
        pygame.draw.circle(pygame.display.get_surface(), self.color, (self.position.x, self.position.y), self.radius, self.width)
        
    def update(self):
        global state
            
        if not state == State.SERVING:
            self.velocity.y += GRAVITY
        self.velocity.x *= drag_factor
        self.velocity.y *= drag_factor
        
        if math.fabs(self.velocity.y) < 0.3:
            self.velocity.y = 0
        
        self.position += self.velocity
        
        if (self.position.x - self.radius) < 0:
            self.velocity.x *= -1
            self.position.x = self.radius
            
        if (self.position.x + self.radius) > SCREEN_WIDTH:
            self.velocity.x *= -1
            self.position.x = SCREEN_WIDTH - self.radius
            
        if (self.position.y - self.radius) < 0 :
            self.position.y = self.radius
            self.velocity.y = 0
            
        if (self.position.y + self.radius) > GROUND_LEVEL:
            # Hit the ground
            if state == State.PRACTICE:
                self.velocity.y *= -bounce_factor
                self.velocity.x *= bounce_factor
                self.position.y = GROUND_LEVEL - self.radius
            else:
                if (self.position.x > SCREEN_WIDTH/2):
                    global left_team_score
                    left_team_score += 1
                    self.position = Vector2(250, 200)
                else:
                    global right_team_score
                    right_team_score += 1
                    self.position = Vector2(SCREEN_WIDTH - 250, 200)
                self.velocity = Vector2(0,0)
                state = State.SERVING
                

class Paddle(GameObject):
    """ A Paddle """
    
    
    def __init__(self, deviceId):
        self.position = Vector2(0,0)
        self.velocity = Vector2(0,0)
        self.position.x, self.position.y = pygame.mouse.get_pos()
        self.velocity.x, self.velocity.y = pygame.mouse.get_rel()
        self.radius = 40
        self.deviceId = deviceId
        self.bodyFrame = None
        red = random.randint(0,255)
        green = random.randint(0,255)
        blue = random.randint(0,255)
        self.color = (red, green, blue)
        
        self.button1Pressed = False
        self.button1Released = False
        self.button2Pressed = False
        self.button2Released = False
        
        self.button1Down = False
        self.button2Down = False
        self.side = Side.LEFT
        
    def draw(self):
        pygame.draw.circle(pygame.display.get_surface(), self.color, (self.position.x,self.position.y), self.radius)
        pygame.draw.circle(pygame.display.get_surface(), BLACK, (self.position.x, self.position.y), self.radius, 4)
        
    def update(self):
        self.bodyFrame = getBodyFrameFromInputThread(self.deviceId)
        if self.bodyFrame is None:
            return
        self.position.x += self.bodyFrame.deltaX
        self.position.y += self.bodyFrame.deltaY
        self.velocity.x = self.bodyFrame.deltaX
        self.velocity.y = self.bodyFrame.deltaY
        
        
        left = 0
        right = SCREEN_WIDTH
        if state == State.PLAY_GAME  or state == State.SERVING   :
            if (self.side == Side.LEFT):
                right = SCREEN_WIDTH/2 - 10 - self.radius
            if (self.side == Side.RIGHT):
                left = SCREEN_WIDTH/2 + 10 + self.radius
        
        if self.position.x < left:
            self.position.x = left
        if self.position.x > right:
            self.position.x = right
        if self.position.y < 0:
            self.position.y = 0
        if self.position.y > SCREEN_HEIGHT:
            self.position.y = SCREEN_HEIGHT
            
            
        self.button1Pressed = self.bodyFrame.button1 and not self.button1Down
        self.button1Released = self.button1Down and not self.bodyFrame.button1
        
        self.button2Pressed = self.bodyFrame.button2 and not self.button2Down
        self.button2Released = self.button2Down and not self.bodyFrame.button2
        
        self.button1Down = self.bodyFrame.button1
        self.button2Down = self.bodyFrame.button2
        
        
        
    def collide(self, ball):
        global state
        if (self.position - ball.position).len() < self.radius + ball.radius:
            pushDir = ball.position - self.position
            
            relativeVelocity = ball.velocity - self.velocity
            backwards = -relativeVelocity
            
            newVel = backwards.reflect(pushDir)
            
            ball.velocity = newVel + self.velocity
            ball.velocity.scale(bounce_factor*bounce_factor)
            
            pushDir = ball.position - self.position
            pushDir.normalize()
            amt = Vector2.dot(self.velocity, pushDir)
            pushDir.scale(amt * bounce_factor)
            ball.velocity += pushDir
            
            if ball.velocity.len() > MAX_SPEED:
                ball.velocity.normalize()
                ball.velocity.scale(MAX_SPEED)
                
            # Find the overlap and move the ball
            moveDir = ball.position - self.position
            moveDir.normalize()
            moveDir *= (ball.radius + self.radius)
            ball.position = self.position + moveDir
            
            if state == State.SERVING:
                state = State.PLAY_GAME

class RectPaddle(Paddle):    
    def draw(self):
        pygame.draw.circle(pygame.display.get_surface(), self.color, (self.position.x,self.position.y), self.radius)
        pygame.draw.circle(pygame.display.get_surface(), BLACK, (self.position.x, self.position.y), self.radius, 4)
        
    

# ------------------ Game static values ------------------------

redball = Ball(SCREEN_WIDTH/2 + 40, 50, RED, 50)
#630, 400, 20, 260
fence = Fence(20, 260, 400)

paddles = []
paddleIds = []

buttons = []

thread = InputThread()

# ---------------------- Buttons -------------------------------

def newGamePressed(sender):
    global state
    state = State.CHOOSE_SIDES
    global buttons
    buttons = []
    leftSideButton = Button(Rect(20,20,(SCREEN_WIDTH/2) - 40, SCREEN_HEIGHT - 140), leftSideCallback)
    rightSideButton = Button(Rect((SCREEN_WIDTH/2) + 20, 20,(SCREEN_WIDTH/2) - 40, SCREEN_HEIGHT - 140), rightSideCallback)
    buttons.append(leftSideButton)
    buttons.append(rightSideButton)
    startGameButton = TextButton(Rect(SCREEN_WIDTH/2 - 50, SCREEN_HEIGHT - 100, 10, 10), startGame, "Start Game")
    buttons.append(startGameButton)

def practicePressed(sender):
    global state
    state = State.PRACTICE

def leftSideCallback(sender):
    sender.side = Side.LEFT
    print sender.deviceId, "chose left side"
    
def rightSideCallback(sender):
    sender.side = Side.RIGHT
    print sender.deviceId, "chose right side"
    
    
def startGame(sender):
    global buttons
    buttons = []
    global state
    state = State.PLAY_GAME
    


# -------------------- Utility Functions -----------------------

def init():
    thread.start()
    pygame.mouse.set_visible(False)
    global font
    font = pygame.font.Font(pygame.font.get_default_font(), 36)
        
def init_update():
    deviceIds = getDeviceIds()
    for id in deviceIds:
        if id not in paddleIds:
            # Make a new paddle
            p = Paddle(id)
            paddles.append(p)
            paddleIds.append(id)
            print "adding new paddle", id
            
    startButton = TextButton(Rect(100,100,50,50), newGamePressed, "New Game")
    buttons.append(startButton)
    practiceButton = TextButton(Rect(100,150,50,50), practicePressed, "Practice")
    buttons.append(practiceButton)
    
    global state
    state = State.MENU

def init_draw():
    screen.fill(BLACK)
    
def game_update():
    global state
    redball.update()
    fence.update()
    for paddle in paddles:
        paddle.update()
        if paddle.button2Pressed:
            state = State.INIT
        paddle.collide(redball)
    fence.collide(redball)

def game_draw():
    screen.fill((0,153,255))
    screen.fill(GREEN, rect=Rect(0, GROUND_LEVEL, 1280, 220))
    redball.draw()
    fence.draw()
    for paddle in paddles:
        paddle.draw()
    
    leftScoreText = left_team_score.__str__()
    rightScoreText = right_team_score.__str__()
    lTextSurface = font.render(leftScoreText, True, WHITE)
    rTextSurface = font.render(rightScoreText, True, WHITE)
    default = pygame.display.get_surface()
    default.blit(lTextSurface, Rect(100,30, 20, 20))
    default.blit(rTextSurface, Rect(1160,30, 20, 20))

def menu_update():
    deviceIds = getDeviceIds()
    for id in deviceIds:
        if id not in paddleIds:
            # Make a new paddle
            p = Paddle(id)
            paddles.append(p)
            paddleIds.append(id)
            print "adding new paddle", id
    for paddle in paddles:
        paddle.update()
    for button in buttons:
        button.update(paddles)
    
def menu_draw():
    screen.fill(BLUE)
    for paddle in paddles:
        paddle.draw()
    for button in buttons:
        button.draw()


def choose_sides_update():
    deviceIds = getDeviceIds()
    for id in deviceIds:
        if id not in paddleIds:
            # Make a new paddle
            p = Paddle(id)
            paddles.append(p)
            paddleIds.append(id)
            print "adding new paddle", id
    for paddle in paddles:
        paddle.update()
    for button in buttons:
        button.update(paddles)
    
def choose_sides_draw():
    screen.fill(GREEN)
    for paddle in paddles:
        paddle.draw()
    for button in buttons:
        button.draw()

def update():
    { State.INIT      : init_update,
      State.PLAY_GAME : game_update,
      State.PRACTICE  : game_update,
      State.MENU      : menu_update,
      State.CHOOSE_SIDES : choose_sides_update,
      State.SERVING   : game_update
     } [state]()
    
def draw():
    { State.INIT      : init_draw,
      State.PLAY_GAME : game_draw,
      State.PRACTICE  : game_draw,
      State.MENU      : menu_draw,
      State.CHOOSE_SIDES : choose_sides_draw,
      State.SERVING   : game_draw
     } [state]()
    pygame.display.flip()


def main():
    try:
        init()
        
        clock = pygame.time.Clock()
    
        while(1):
            clock.tick(60)
            # The game loop
            for event in pygame.event.get():
                if event.type == QUIT:
                    return
    
            update()
    
            draw()
    finally:
        stopInputThread()
        thread.join()
        

# ---------------------- Main! --------------------------------

if (__name__ == "__main__"):
    pygame.init()
    main()
    pygame.quit()
    
