#!/usr/bin/env python
#
# This file is part of pyfreespace.
#  
# Copyright (c) 2009, Hillcrest Laboratories, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     * Neither the name of the Hillcrest Laboratories, Inc. nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import freespace
from sys import platform


if platform == 'win32':
    import win32event;
    rlist = []
else:
    from select import poll
    poll = poll()


def add_pollfd(fd, events):
    if platform == 'win32':
        rlist.append(fd)
    else:
        poll.register(fd, events)


def remove_pollfd(fd):
    if platform == 'win32':
        rlist.remove(fd)
    else:
        poll.unregister(fd)


def receiveCallback(device, message, rc):
    if rc == freespace.FREESPACE_SUCCESS:
        print '%d> %s' % (device.id, message.__str__())
    else:
        print "Error %d." % rc
        if rc == freespace.FREESPACE_ERROR_NOT_FOUND:
            device.close();


def initDevice(device):
    # --- START EXAMPLE INITIALIZATION OF DEVICE -- 
    
    print "Found a device. Trying to open it..."
    device.open()

    # Display the device information.
    print device
    
    device.set_receive_callback(receiveCallback);

    print "Opened. Flushing any queued messages."
    device.flush()

    print "Sending message to enable body-frame motion data."
    d = freespace.DataMotionControl()
    d.enableBodyMotion = True;

    device.sendmessage_async(d, 1000)
    print "Device %d has been initialized" % device.id
    # --- END EXAMPLE INITIALIZATION OF DEVICE -- 


def cleanupDevice(device):
    # --- START EXAMPLE FINALIZATION OF DEVICE ---
    print "%d> Sending message to enable mouse motion data." % device.id
    d = freespace.DataMotionControl()
    d.enableMouseMovement = True;
    
    rc = device.sendmessage(d)
    
    print "%d> Cleaning up..." % device.id
    device.close()
    # --- END EXAMPLE FINALIZATION OF DEVICE ---
    

def hotplugCallback(event, device):
    if event == freespace.FREESPACE_HOTPLUG_REMOVAL:
        print "Closing removed device %d" % device.id
        device.close();
    elif event == freespace.FREESPACE_HOTPLUG_INSERTION:
        print "Opening newly inserted device %d" % device.id
        initDevice(device);


# ------------- Main ----------------

def runMain():
    # Initialize the freespace library
    freespace.init();

    freespace.set_hotplug_callback(hotplugCallback)
    freespace.set_filedescriptor_callbacks(add_pollfd, remove_pollfd)
    freespace.sync_filedescriptors()

    freespace.perform()
    
    try:
        while (True) :
            timeoutMs = freespace.get_next_timeout();
    
            # Adjust the maximum timeout to allow for Ctrl-C handling
            #if (timeoutMs < 0 or timeoutMs > 1000) :
            #    timeoutMs = 1000;
            
            #print timeoutMs
            if platform == 'win32':
                #(outrlist, foo, bar) = select.select(rlist, [], [], timeoutMs)
                win32event.WaitForMultipleObjects(rlist, 0, timeoutMs)
            else:
                events = poll.poll(timeoutMs)
                if len(events) == 0:
                    continue
            freespace.perform();
    except KeyboardInterrupt:
        print "Caught Keyboard Interrupt"
    finally:
        print "Cleaning up all devices..."
        
        list = freespace.get_device_list();
        for device in list:
            cleanupDevice(device)
    
        freespace.exit()

# ---------------------- Main! --------------------------------

if (__name__ == "__main__"):
    runMain()
    
